# Copyright(c) 2002-2004 Association of Universities for Research in Astronomy, Inc.
#
# GMOS example reductions script: Typical reduction of spectrophomotometric 
#                                 standard star observation
#
# This data processing was done to aid the data quality assessment performed
# by the Gemini staff. The data processing is not designed to give the best
# possible result. Better signal-to-noise and better cleaning for cosmic-ray
# hits and bad pixels will most likely be possible. The user of these data
# is encouraged to use the provided co-added images only as guide lines and
# to re-reduce the data to obtain the best possible reduction.
#
# Gemini GMOS data reduction script
# Observation UT date: 2001dec22 
# Data processor:   Inger Jorgensen
# Data reduction date: 2002feb25
#
# Brief data description: Spectrophomotometric standard star observation
# 
# Define various directories containing raw data and calibration data
set raw=/net/sabrina/staging2/gmos/2001dec22/
set eraw=/net/sabrina/staging2/gmos/2001dec23/
set gcalib=../../2001dec25/Basecalib/

# set up the logfile for this reduction
gmos.logfile="GN-CAL20011222_log.txt"

# The bias image has been made with gbias

# Make the flat field
gsflat N20011221S128.fits rgN20011221S128_flat.fits order=29 \
rawpath=raw$ bias=gcalib$N20011217S182_bias.fits 

# In this wavelength range the flats contain a pair of emission lines
# (due to a yet unidentified emission line source in the ISS)
# Set the pixels in those lines to one to avoid introducing spurious
# features
imreplace rgN20011221S128_flat.fits[sci,1][2350:2425,*] 1

# Reduce the observations of the star, the name of the output image is 
# defined by the default value outpref="gs"
# Steps in gsreduce 
#    subtract off the bias 
#    mosaic the 3 detectors 
#    interpolate accross the chip gaps to aid later reduction steps
#    flat field correction 
#    run gsappwave to get an approximate wavelength calibration in the 
#       header of the output image
#
gsreduce N20011221S129 rawpath=raw$ \
bias=gcalib$N20011217S182_bias.fits flat=rgN20011221S128_flat.fits

# CuAr B600/660 from the night after, since no arcs were taken 2001dec22
# The CuAr is not flat fielded, gaps are not fixpixed
gsreduce N20011222S027 rawpath=eraw$ fl_flat- fl_fixpix- \
bias=gcalib$N20011217S182_bias.fits 

# Establish the wavelength calibration
gswavelength gsN20011222S027.fits coordlist="linelists$cuar.dat" \
fwidth=10 cradius=12 minsep=5

# Transform the CuAr spectrum, for checking that the transformation is ok.
# Output image name is defined by the default value outpref="t"
gstransform gsN20011222S027 wavtran=gsN20011222S027
# Transform the star spectrum
gstransform gsN20011221S129 wavtran=gsN20011222S027

# Skysubtract the star spectrum. Output image name is defined by the
# default value outpref="s"
gsskysub tgsN20011221S129 long_sample="400:500,600:700"

# Extract the 1D spectrum (central 1arcsec)
gsextract stgsN20011221S129 

# Extablish the sensitivity function, the standard star is G191B2B
gsstandard estgsN20011221S129 std sens starname=G191B2B

# Apply the calibration to the extracted 1D spectrum
# The default value sfunction = "sens" defines the sensitivity function
# created above as the calibration file
gscalibrate estgsN20011221S129 

# Apply the calibration to the 2D spectrum
# The default value sfunction = "sens" defines the sensitivity function
# created above as the calibration file
gscalibrate stgsN20011221S129 

